@@ -14,7 +14,8 @@ use Tie::ToObject;
 
 __PACKAGE__->mk_accessors(qw(tied_as_objects));
 
-use constant DEBUG => our $DEBUG || $ENV{DATA_VISITOR_DEBUG};
+# the double not works makes this no longer undef, so exempt from useless constant warnings in older perls
+use constant DEBUG => not not our $DEBUG || $ENV{DATA_VISITOR_DEBUG};
 
 sub trace {
 	my ( $self, $category, @msg ) = @_;
@@ -34,7 +35,7 @@ sub _print_trace {
 	warn "@msg\n";
 }
 
-our $VERSION = "0.14";
+our $VERSION = "0.16";
 
 sub visit {
 	my ( $self, $data ) = @_;
@@ -46,7 +47,7 @@ sub visit {
 	if ( ref $data ) { # only references need recursion checks
 		if ( exists $seen_hash->{ refaddr($data) } ) {
 			$self->trace( mapping => found_mapping => from => $data, to => $seen_hash->{ refaddr($data) } ) if DEBUG;
-			return $seen_hash->{ refaddr($data) }; # return whatever it was mapped to
+			return $self->visit_seen( $data, $seen_hash->{refaddr($data)} );
 		} else {
 			$self->trace( mapping => no_mapping => $data ) if DEBUG;
 		}
@@ -55,6 +56,11 @@ sub visit {
 	return $self->visit_no_rec_check( $data );
 }
 
+sub visit_seen {
+	my ( $self, $data, $result ) = @_;
+	return $result;
+}
+
 sub _get_mapping {
 	my ( $self, $data ) = @_;
 	$self->{_seen}{ refaddr($data) };
@@ -96,12 +102,13 @@ sub visit_ref {
 
 	$reftype = "SCALAR" if $reftype =~ /^(?:REF|LVALUE|VSTRING)$/;
 
-	my $method = lc "visit_$reftype";
+	my $method = $self->can(lc "visit_$reftype") || "visit_value";
 
-	if ( $self->can($method) ) {
-		return $self->_register_mapping( $data, $self->$method($data) );
+	if ( not defined wantarray ) {
+		$self->_register_mapping( $data, $data );
+		return $self->$method($data);
 	} else {
-		return $self->_register_mapping( $data, $self->visit_value($data) );
+		return $self->_register_mapping( $data, $self->$method($data) );
 	}
 
 }
@@ -120,6 +127,7 @@ sub visit_hash {
 	if ( not defined wantarray ) {
 		$self->_register_mapping( $hash, $hash );
 		$self->visit_hash_entries($hash);
+		return;
 	} else {
 		my $new_hash = {};
 		$self->_register_mapping( $hash, $new_hash );
@@ -138,8 +146,12 @@ sub visit_hash {
 
 sub visit_hash_entries {
 	my ( $self, $hash ) = @_;
-	no warnings 'void';
-	map { $self->visit_hash_entry( $_, $hash->{$_}, $hash ) } keys %$hash;
+
+	if ( not defined wantarray ) {
+		$self->visit_hash_entry( $_, $hash->{$_}, $hash ) for keys %$hash;
+	} else {
+		return map { $self->visit_hash_entry( $_, $hash->{$_}, $hash ) } keys %$hash;
+	}
 }
 
 sub visit_hash_entry {
@@ -147,10 +159,15 @@ sub visit_hash_entry {
 
 	$self->trace( flow => visit_hash_entry => key => $key, value => $value ) if DEBUG;
 
-	return (
-		$self->visit_hash_key($key,$value,$hash),
-		$self->visit_hash_value($_[2],$key,$hash) # retain aliasing semantics
-	);
+	if ( not defined wantarray ) {
+		$self->visit_hash_key($key,$value,$hash);
+		$self->visit_hash_value($_[2],$key,$hash); # retain aliasing semantics
+	} else {
+		return (
+			$self->visit_hash_key($key,$value,$hash),
+			$self->visit_hash_value($_[2],$key,$hash) # retain aliasing semantics
+		);
+	}
 }
 
 sub visit_hash_key {
@@ -169,7 +186,7 @@ sub visit_array {
 	if ( not defined wantarray ) {
 		$self->_register_mapping( $array, $array );
 		$self->visit_array_entries($array);
-		$self->visit_array_entry( $array->[$_], $_, $array ) for 0 .. $#$array
+		return;
 	} else {
 		my $new_array = [];
 		$self->_register_mapping( $array, $new_array );
@@ -188,8 +205,12 @@ sub visit_array {
 
 sub visit_array_entries {
 	my ( $self, $array ) = @_;
-	no warnings 'void';
-	map { $self->visit_array_entry( $array->[$_], $_, $array ) } 0 .. $#$array;
+
+	if ( not defined wantarray ) {
+		$self->visit_array_entry( $array->[$_], $_, $array ) for 0 .. $#$array;
+	} else {
+		return map { $self->visit_array_entry( $array->[$_], $_, $array ) } 0 .. $#$array;
+	}
 }
 
 sub visit_array_entry {
