@@ -6,11 +6,13 @@ use base qw/Data::Visitor/;
 use strict;
 use warnings;
 
+use Carp qw(carp);
 use Scalar::Util qw/blessed refaddr reftype/;
 
 __PACKAGE__->mk_accessors( qw/callbacks class_callbacks ignore_return_values/ );
 
-BEGIN { *DEBUG = \&Data::Visitor::DEBUG }
+use constant DEBUG => Data::Visitor::DEBUG();
+use constant FIVE_EIGHT => ( $] >= 5.008 );
 
 sub new {
 	my ( $class, %callbacks ) = @_;
@@ -25,7 +27,22 @@ sub new {
 		$tied_as_objects = delete $callbacks{tied_as_objects};
 	}
 
-	my @class_callbacks = do { no strict 'refs'; grep { defined %{"${_}::"} } keys %callbacks };
+	my @class_callbacks = do {
+		no strict 'refs';
+		grep {
+			# this check can be half assed because an ->isa check will be
+			# performed later. Anything that cold plausibly be a class name
+			# should be included in the list, even if the class doesn't
+			# actually exist.
+
+			m{ :: | ^[A-Z] }x # if it looks kinda lack a class name
+				or
+			scalar keys %{"${_}::"} # or it really is a class
+		} keys %callbacks;
+	};
+
+	# sort from least derived to most derived
+	@class_callbacks = sort { !$a->isa($b) <=> !$b->isa($a) } @class_callbacks;
 
 	$class->SUPER::new({
 		tied_as_objects => $tied_as_objects,
@@ -43,14 +60,31 @@ sub visit {
 	local *_ = \$_[1]; # alias $_
 
 	if ( ref $data and exists $replaced_hash->{ refaddr($data) } ) {
-		$self->trace( mapping => replace => $data, with => $replaced_hash->{ refaddr($data) } ) if DEBUG;
-		return $_[1] = $replaced_hash->{ refaddr($data) };
+		if ( FIVE_EIGHT ) {
+			$self->trace( mapping => replace => $data, with => $replaced_hash->{ refaddr($data) } ) if DEBUG;
+			return $_[1] = $replaced_hash->{ refaddr($data) };
+		} else {
+			carp(q{Assignment of replacement value for already seen reference } . overload::StrVal($data) . q{ to container doesn't work on Perls older than 5.8, structure shape may have lost integrity.});
+		}
+	}
+
+	my $ret = $self->SUPER::visit( $self->callback( visit => $data ) );
+
+	$replaced_hash->{ refaddr($data) } = $_ if ref $data and ( not ref $_ or refaddr($data) ne refaddr($_) );
+
+	return $ret;
+}
+
+sub visit_seen {
+	my ( $self, $data, $result ) = @_;
+
+	my $mapped = $self->callback( seen => $data, $result );
+
+	no warnings 'uninitialized';
+	if ( refaddr($mapped) == refaddr($data) ) {
+		return $result;
 	} else {
-		my $ret = $self->SUPER::visit( $self->callback( visit => $data ) );
-
-		$replaced_hash->{ refaddr($data) } = $_ if ref $data and ( not ref $_ or refaddr($data) ne refaddr($_) );
-
-		return $ret;
+		return $mapped;
 	}
 }
 
@@ -87,7 +121,7 @@ sub visit_object {
 	$data;
 }
 
-sub subname { $_[2] }
+sub subname { $_[1] }
 
 BEGIN {
 	eval {
